import re
from typing import Dict

from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.

    Note: In schema version 10.3.0, the `NomAnalysisActivity.id` slot had this `structured_pattern.syntax`:
          "{id_nmdc_prefix}:wfnom-{id_shoulder}-{id_blade}{id_version}{id_locus}"

          In schema version 10.4.0, it has _this_ `structured_pattern.syntax`:
          "{id_nmdc_prefix}:wfnom-{id_shoulder}-{id_blade}{id_version}"

          That specific change was made in commit `a1d5686`:
          https://github.com/microbiomedata/nmdc-schema/commit/a1d5686

          This migrator was designed to migrate `NomAnalysisActivity.id` values to conform to the latter pattern,
          and to update the `was_generated_by` fields of the associated `data_object_set` documents accordingly.
    """

    _from_version = "10.3.0"
    _to_version = "10.4.0"

    # Note: I think the author of the migrator got this regex pattern by running `$ make all`
    #       and getting this pattern from the file `project/nmdc_materialized_patterns.yaml`.
    #       Reference: https://github.com/microbiomedata/nmdc-schema/pull/2059#discussion_r1639005255
    #
    pattern = re.compile(
        r"(^(nmdc):wfnom-([0-9][a-z]{0,6}[0-9])-([A-Za-z0-9]{1,})(\.[0-9]{1,})$)"
    )

    # A mapping from `DataObject.id` value to `NomAnalysisActivity.id` value.
    data_object_id_to_nom_id_map: Dict[str, str] = dict()

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document(
            "nom_analysis_activity_set",
            [
                self.populate_alternative_identifiers,
                self.update_nom_analysis_activity_id,
            ],
        )

        self.adapter.process_each_document(collection_name="data_object_set",
                                           pipeline=[self.update_was_generated_by_field])

    def update_was_generated_by_field(self, data_object: dict) -> dict:
        r"""
        Updates the `was_generated_by` field of the specified `data_object_set` document so that it contains
        the `id` of the `nom_analysis_activity_set` document that "generated" that `data_object_set` document.

        >>> m = Migrator()
        >>> m.data_object_id_to_nom_id_map = {'do1': 'nom1.1'}
        >>> m.update_was_generated_by_field({'id': 'do1', 'was_generated_by': 'nom1'})  # changes `was_generated_by`
        {'id': 'do1', 'was_generated_by': 'nom1.1'}
        >>> m.update_was_generated_by_field({'id': 'do2', 'was_generated_by': 'nom2'})  # no change (not in dictionary)
        {'id': 'do2', 'was_generated_by': 'nom2'}
        """

        data_object_id = data_object["id"]
        if data_object_id in self.data_object_id_to_nom_id_map.keys():
            data_object["was_generated_by"] = self.data_object_id_to_nom_id_map[data_object_id]

        return data_object

    def is_valid_id(self, s: str) -> bool:
        r"""
        Helper function that checks whether a string is a valid ID.
        Reference: https://docs.python.org/3/library/re.html#re.Pattern.search

        >>> m = Migrator()
        >>> m.is_valid_id("nmdc:wfnom-13-7yf9qj85")
        False
        >>> m.is_valid_id("nmdc:wfnom-13-7yf9qj85.1")
        True
        >>> m.is_valid_id("nmdc:wfnom-13-7yf9qj85.10")
        True
        """

        return self.pattern.search(s) is not None

    def populate_alternative_identifiers(self, nom_analysis_activity: dict) -> dict:
        r"""
        Checks to see if a nom_analysis_activity has any alternative_identifiers.
        If there are any, then the ID is added if it's invalid.
        If there are not any, then the alternative_identifiers slot is created
        if needed with the invalid id as a value.

        >>> m = Migrator()
        >>> # test: adds alternative identifiers
        >>> m.populate_alternative_identifiers({'id': 'nmdc:wfnom-13-7yf9qj85'})
        {'id': 'nmdc:wfnom-13-7yf9qj85', 'alternative_identifiers': ['nmdc:wfnom-13-7yf9qj85']}
        >>> # test: no change, alternative identifiers already present with valid id
        >>> m.populate_alternative_identifiers({'id': 'nmdc:wfnom-13-7yf9qj85.1',
        ...                                     'alternative_identifiers': ['alt 1']})
        {'id': 'nmdc:wfnom-13-7yf9qj85.1', 'alternative_identifiers': ['alt 1']}
        >>> # test: no change, alternative identifiers not present, but with valid id
        >>> m.populate_alternative_identifiers({'id': 'nmdc:wfnom-13-7yf9qj85.1'})
        {'id': 'nmdc:wfnom-13-7yf9qj85.1'}

        """
        self.logger.info(
            f"Checking alternative_identifiers for NomAnalysisActivity: {nom_analysis_activity['id']}"
        )

        nom_id = nom_analysis_activity["id"]
        if not self.is_valid_id(nom_id):
            if "alternative_identifiers" not in nom_analysis_activity.keys():
                nom_analysis_activity["alternative_identifiers"] = []
            nom_analysis_activity["alternative_identifiers"].append(nom_id)

        return nom_analysis_activity

    def update_nom_analysis_activity_id(self, nom_analysis_activity: dict) -> dict:
        r"""
        If the specified `nom_analysis_activity` document has an invalid `id`,
        this function does two things:
        1. Make the `id` valid (by appending a version number to it); and
        2. If it has a `has_output` field (which is a list) consisting of a single item,
           add an entry to the `data_object_id_to_nom_id_map` dictionary, using
           the `has_output` value as the key and the newly-valid `id` as the value.
           Raise an exception if the `has_output` list contains more than 1 item.

        If the specified `nom_analysis_activity` document already has a valid `id`,
        this function does nothing.

        >>> m = Migrator()
        >>> # test: adds version number to id
        >>> m.update_nom_analysis_activity_id({'id': 'nmdc:wfnom-13-7yf9qj85'})
        {'id': 'nmdc:wfnom-13-7yf9qj85.1'}
        >>> # test: if version number is already there, no change
        >>> m.update_nom_analysis_activity_id({'id': 'nmdc:wfnom-13-7yf9qj85.1'})
        {'id': 'nmdc:wfnom-13-7yf9qj85.1'}

        Test the production of the ID map.

        >>> m.data_object_id_to_nom_id_map = {}
        >>> m.update_nom_analysis_activity_id({'id': 'nmdc:wfnom-13-7yf9qj85',
        ...                                    'has_output': ['do1']})  # creates dict entry
        {'id': 'nmdc:wfnom-13-7yf9qj85.1', 'has_output': ['do1']}
        >>> len(m.data_object_id_to_nom_id_map.keys())
        1
        >>> m.data_object_id_to_nom_id_map['do1']
        'nmdc:wfnom-13-7yf9qj85.1'
        >>> m.update_nom_analysis_activity_id({'id': 'nmdc:wfnom-13-7yf9qj86',
        ...                                    'has_output': ['do2']})  # creates other dict entry
        {'id': 'nmdc:wfnom-13-7yf9qj86.1', 'has_output': ['do2']}
        >>> len(m.data_object_id_to_nom_id_map.keys())
        2
        >>> m.data_object_id_to_nom_id_map['do1']
        'nmdc:wfnom-13-7yf9qj85.1'
        >>> m.data_object_id_to_nom_id_map['do2']
        'nmdc:wfnom-13-7yf9qj86.1'
        """
        nom_analysis_activity_id = nom_analysis_activity["id"]

        self.logger.info(
            f"Checking ID for NomAnalysisActivity: {nom_analysis_activity_id}"
        )

        # Get the `has_output` value, if any, of the `nom_analysis_activity` document;
        # treating `null` values as empty lists.
        nom_outputs = []
        if "has_output" in nom_analysis_activity.keys():
            if nom_analysis_activity["has_output"] is not None:
                nom_outputs = nom_analysis_activity["has_output"]

        # If the `id` is invalid, do two things:
        # 1. Append a version number to it; and
        # 2. Record the `id` of the `DataObject`, if any, that is the output of this `NomAnalysisActivity`.
        if not self.is_valid_id(nom_analysis_activity_id):
            nom_analysis_activity_id_new = f"{nom_analysis_activity_id}.1"
            nom_analysis_activity["id"] = nom_analysis_activity_id_new

            if len(nom_outputs) > 1:
                raise ValueError(f"NomAnalysisActivity ({nom_analysis_activity_id}) 'has_output' field "
                                 f"contains more than one item. Contains: {nom_outputs}")
            elif len(nom_outputs) == 1:
                data_object_id = nom_outputs[0]
                if data_object_id in self.data_object_id_to_nom_id_map.keys():
                    raise ValueError(f"Multiple NomAnalysisActivity instances have same DataObject ({data_object_id}) "
                                     f"instance as their output.")
                self.data_object_id_to_nom_id_map[data_object_id] = nom_analysis_activity_id_new
            else:  # This `NomAnalysisActivity` instance didn't have any output.
                pass

        return nom_analysis_activity



from linkml_runtime import SchemaView

schema_file = "../src/schema/nmdc.yaml"

schema_view = SchemaView(schema_file)

schema_classes = schema_view.all_classes()

for ck, _ in schema_classes.items():
    induced_slots = schema_view.class_induced_slots(ck)
    for induced_slot in induced_slots:
        if induced_slot.name == 'used':
            print(f"In class {ck}, slot used has range  {induced_slot.range}")



import unittest
from typing import Optional

# TODO: pymongo is used here, but it is not listed as a production dependency of nmdc-schema (only as a dev-dependency).
from pymongo import MongoClient, timeout
from pymongo.database import Database

from nmdc_schema.migrators.adapters.mongo_adapter import MongoAdapter


# Define constants used by the test.
# TODO: Move these to an environment configuration file (is there a standard one; e.g. "local/.env")?
MONGO_HOST: str = "host.docker.internal"
MONGO_PORT: int = 27017
MONGO_USER: Optional[str] = None
MONGO_PASS: Optional[str] = None
MONGO_DATABASE_NAME: str = "test-nmdc-schema"
MONGO_TIMEOUT_DURATION: int = 3  # seconds


class TestMongoAdapter(unittest.TestCase):
    r"""
    Tests targeting the `MongoAdapter` class.

    You can start up a containerized MongoDB server like this:
    $ docker run --rm --detach --name mongo-test-nmdc-schema -p 27017:27017 mongo

    One that's running, other containers will be able to access it via:
    - host.docker.internal:27017

    You can run these tests like this:
    $ poetry run python -m unittest -v nmdc_schema/migrators/adapters/test_mongo_adapter.py

    References:
    - https://docs.python.org/3/library/unittest.html#basic-example
    - https://docs.python.org/3/library/unittest.html#unittest.TestCase.setUp
    """

    mongo_client: Optional[MongoClient] = None
    db: Optional[Database] = None

    def setUp(self) -> None:
        r"""
        Connects to the MongoDB server and gets a reference to the database.

        Note: This function runs before each test starts.
        """

        # Connect to the MongoDB server and store a reference to the connection.
        self.mongo_client = MongoClient(
            host=MONGO_HOST,
            port=MONGO_PORT,
            username=MONGO_USER,
            password=MONGO_PASS,
            authSource="admin",
        )
        with timeout(MONGO_TIMEOUT_DURATION):
            # Try connecting to the database server.
            _ = self.mongo_client.server_info()
            db = self.mongo_client[MONGO_DATABASE_NAME]

            # Ensure the database contains no collections.
            if len(db.list_collection_names()):
                raise KeyError(f"Database is not empty: {MONGO_DATABASE_NAME}")

        # Store a reference to the database.
        self.db = db

    def tearDown(self) -> None:
        r"""
        Drops all collections in the database and closes the connection to the MongoDB server.

        Note: This function runs after each test finishes.
        """

        # Drop all collections in the database.
        for collection_name in self.db.list_collection_names():
            self.db.drop_collection(collection_name)

        # Close the connection to the server.
        self.mongo_client.close()

    def test_create_collection(self):
        # Set up:
        collection_name = "my_collection"

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.create_collection(collection_name)

        # Validate result:
        collection_names_after = self.db.list_collection_names()
        assert len(collection_names_after) == 1
        assert collection_name in collection_names_after

    def test_rename_collection(self):
        # Set up:
        collection_name_original = "my_collection"
        collection_name_target = "your_collection"
        self.db.create_collection(collection_name_original)
        assert len(self.db.list_collection_names()) == 1

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.rename_collection(collection_name_original, collection_name_target)

        # Validate result:
        collection_names_after = self.db.list_collection_names()
        assert len(collection_names_after) == 1
        assert collection_name_target == collection_names_after[0]

    def test_delete_collection(self):
        # Set up:
        collection_name = "my_collection"
        self.db.create_collection(collection_name)
        assert len(self.db.list_collection_names()) == 1

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.delete_collection(collection_name)

        # Validate result:
        assert len(self.db.list_collection_names()) == 0

    def test_insert_document(self):
        # Set up:
        collection_name = "my_collection"
        document = dict(id=123, foo="bar")
        self.db.create_collection(collection_name)
        assert self.db.get_collection(collection_name).count_documents({}) == 0

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.insert_document(collection_name, document)

        # Validate result:
        assert self.db.get_collection(collection_name).count_documents({}) == 1
        document_actual = self.db.get_collection(collection_name).find_one({})
        for k, v in document.items():
            assert v == document_actual[k]

    def test_get_document_having_value_in_field(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(id=1, x="a")
        document_2 = dict(id=2, x="b")
        document_3 = dict(id=3, x="c")
        self.db.create_collection(collection_name)
        self.db.get_collection(collection_name).insert_many(
            [document_1, document_2, document_3]
        )
        assert self.db.get_collection(collection_name).count_documents({}) == 3

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        document_actual = adapter.get_document_having_value_in_field(
            collection_name, "x", "b"
        )

        # Validate result:
        for k, v in document_2.items():
            assert v == document_actual[k]

    def test_get_document_having_one_of_values_in_field(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(id=1, x="a")
        document_2 = dict(id=2, x="b")
        document_3 = dict(id=3, x="c")
        self.db.create_collection(collection_name)
        self.db.get_collection(collection_name).insert_many(
            [document_1, document_2, document_3]
        )
        assert self.db.get_collection(collection_name).count_documents({}) == 3

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        document_actual = adapter.get_document_having_one_of_values_in_field(
            collection_name, "x", ["n", "b", "m"]
        )

        # Validate result:
        for k, v in document_2.items():
            assert v == document_actual[k]

    def test_delete_documents_having_value_in_field(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(id=1, x="a")
        document_2 = dict(id=2, x="b")
        document_3 = dict(id=3, x="c")
        document_4 = dict(id=4, x="b")  # note: x="b" here also
        self.db.create_collection(collection_name)
        self.db.get_collection(collection_name).insert_many(
            [document_1, document_2, document_3, document_4]
        )
        assert self.db.get_collection(collection_name).count_documents({}) == 4
        assert self.db.get_collection(collection_name).count_documents({"x": "b"}) == 2

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        num_documents_deleted = adapter.delete_documents_having_value_in_field(
            collection_name, "x", "b"
        )

        # Validate result:
        assert num_documents_deleted == 2
        assert self.db.get_collection(collection_name).count_documents({}) == 2
        assert self.db.get_collection(collection_name).count_documents({"x": "b"}) == 0

    def test_process_each_document(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(_id=1, id=1, x="a")
        document_2 = dict(_id=2, id=2, x="b")
        document_3 = dict(_id=3, id=3, x="c")
        self.db.create_collection(collection_name)
        self.db.get_collection(collection_name).insert_many(
            [document_1, document_2, document_3]
        )

        def capitalize_x(doc: dict) -> dict:
            r"""Example pipeline stage that capitalizes the first letter of the `x` value."""
            doc["x"] = doc["x"].capitalize()
            return doc

        def append_z_to_x_value(doc: dict) -> dict:
            r"""Example pipeline stage that appends "z" to the `x` value."""
            doc["x"] = doc["x"] + "z"
            return doc

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.process_each_document(
            collection_name, [capitalize_x, append_z_to_x_value]
        )

        # Validate result:
        collection = self.db.get_collection(collection_name)
        assert collection.count_documents({"id": 1, "x": "a"}) == 0  # pre-pipeline
        assert collection.count_documents({"id": 1, "x": "b"}) == 0
        assert collection.count_documents({"id": 1, "x": "c"}) == 0
        assert collection.count_documents({"id": 1, "x": "A"}) == 0  # mid-pipeline
        assert collection.count_documents({"id": 1, "x": "B"}) == 0
        assert collection.count_documents({"id": 1, "x": "C"}) == 0
        assert collection.count_documents({"id": 1, "x": "Az"}) == 1  # post-pipeline
        assert collection.count_documents({"id": 2, "x": "Bz"}) == 1
        assert collection.count_documents({"id": 3, "x": "Cz"}) == 1

    def test_set_field_of_each_document(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(_id=1, id=1, x="original")
        document_2 = dict(_id=2, id=2)
        document_3 = dict(_id=3, id=3, x=None)
        self.db.create_collection(collection_name)
        self.db.get_collection(collection_name).insert_many(
            [document_1, document_2, document_3]
        )

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.set_field_of_each_document(collection_name, "x", "new")

        # Validate result:
        collection = self.db[collection_name]
        assert collection.count_documents({"x": "original"}) == 0
        assert collection.count_documents({"x": {"$exists": False}}) == 0
        assert collection.count_documents({"x": None}) == 0
        assert collection.count_documents({"_id": 1, "id": 1, "x": "new"}) == 1
        assert collection.count_documents({"_id": 2, "id": 2, "x": "new"}) == 1
        assert collection.count_documents({"_id": 3, "id": 3, "x": "new"}) == 1

    def test_do_for_each_document(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(_id=1, id=1, x="a")
        document_2 = dict(_id=2, id=2, x="b")
        document_3 = dict(_id=3, id=3, x="c")
        self.db.create_collection(collection_name)
        self.db.get_collection(collection_name).insert_many(
            [document_1, document_2, document_3]
        )
        # Temporarily add an attribute to this class instance so that
        # this test has something persistent it can modify and examine.
        self._characters = []

        def append_x_to_sequence(doc: dict) -> None:
            r"""Example pipeline stage that appends the `x` value to some list."""
            self._characters.append(doc["x"])

        # Invoke function-under-test:
        adapter = MongoAdapter(database=self.db)
        adapter.do_for_each_document(
            collection_name, append_x_to_sequence
        )

        # Validate result:
        # - The list consists of the `x` values from the documents in the collection.
        assert len(self._characters) == 3
        assert self._characters[0] == "a"
        assert self._characters[1] == "b"
        assert self._characters[2] == "c"
        # - The collection was not modified.
        collection = self.db.get_collection(collection_name)
        assert collection.count_documents({"_id": 1, "id": 1, "x": "a"}) == 1
        assert collection.count_documents({"_id": 2, "id": 2, "x": "b"}) == 1
        assert collection.count_documents({"_id": 3, "id": 3, "x": "c"}) == 1

        # Clean up:
        delattr(self, "_characters")

    def test_callbacks(self):
        # Set up:
        collection_name = "my_collection"
        new_collection_name = "my_new_collection"
        event_log: list[str] = []
        adapter = MongoAdapter(
            database=self.db,
            on_collection_created=lambda name: event_log.append(f"Created {name}"),
            on_collection_renamed=lambda old_name, name: event_log.append(
                f"Renamed {old_name} to {name}"
            ),
            on_collection_deleted=lambda name: event_log.append(f"Deleted {name}"),
        )
        assert len(event_log) == 0

        # Trigger callback and validate latest message in event log:
        adapter.create_collection(collection_name)
        assert len(event_log) == 1
        assert event_log[0] == f"Created {collection_name}"

        # Trigger callback and validate latest message in event log:
        adapter.rename_collection(collection_name, new_collection_name)
        assert len(event_log) == 2
        assert event_log[1] == f"Renamed {collection_name} to {new_collection_name}"

        # Trigger callback and validate latest message in event log:
        adapter.delete_collection(new_collection_name)
        assert len(event_log) == 3
        assert event_log[2] == f"Deleted {new_collection_name}"


if __name__ == "__main__":
    unittest.main()



import json
import pymongo
import os
from pymongo import MongoClient
from pymongo.errors import ConnectionFailure
from dotenv import load_dotenv
import urllib

# this is generating a .json data file that can be used to delete these records by submitting it to the appropriate endpoint (ie queries:run)
# for EMP500 (nmdc:sty-11-547rwq94) which have different base IDs with a .1 version.

envfile_path = "../../.env.client"

load_dotenv(envfile_path)
username = os.environ["MONGO_USER"]
pw = os.environ["MONGO_PW"]
host = os.environ["HOST"]
port = os.environ["PORT"]

prod_mongo = (
    "mongodb://"
    + username
    + ":"
    + urllib.parse.quote(pw)
    + "@"
    + host
    + ":"
    + port
    + "/?authMechanism=SCRAM-SHA-256&authSource=admin&directConnection=true"
)

client = MongoClient(prod_mongo)
mydb = client["nmdc"]


study_id = "nmdc:sty-11-547rwq94"

# collection_names
asm_coll_name = "metagenome_assembly_set"
qc_coll_name = "read_qc_analysis_activity_set"
rbt_coll_name = "read_based_taxonomy_analysis_activity_set"
do_coll_name = "data_object_set"
# collection connections
asm_coll = mydb[asm_coll_name]
qc_coll = mydb[qc_coll_name]
rbt_coll = mydb[rbt_coll_name]
do_coll = mydb[do_coll_name]

# aggregation pipeline varaibles
pipeline = [
    {"$group": {"_id": "$was_informed_by", "count": {"$sum": 1}}},
    {"$match": {"count": {"$gt": 1}}},
    {
        "$lookup": {
            "from": "omics_processing_set",
            "localField": "_id",
            "foreignField": "id",
            "as": "omics_processing_set",
        }
    },
    {"$match": {"omics_processing_set.part_of": "nmdc:sty-11-547rwq94"}},
]


# look for the newest assembly record, look up upstream filtering record and reads based taxonomy analysis that uses the same record
def pick_records(omics_id):
    pick_records_pipeline = [
        {"$match": {"was_informed_by": omics_id}},
        {"$sort": {"_id": -1}},
        {"$limit": 1},
        {
            "$lookup": {
                "from": "read_qc_analysis_activity_set",
                "localField": "has_input",
                "foreignField": "has_output",
                "as": "upstream_filtering",
            }
        },
        {
            "$lookup": {
                "from": "read_based_taxonomy_analysis_activity_set",
                "localField": "has_input",
                "foreignField": "has_input",
                "as": "rbt_record_to_keep",
            }
        },
    ]
    records = list(asm_coll.aggregate(pick_records_pipeline))
    return records


# if there are multiple read based taxonomy records, only return the oldest


def pick_rbt(input_do):

    pick_oldest_rbt = [
        {"$match": {"has_input": input_do}},
        {"$sort": {"_id": -1}},
        {"$limit": 1},
    ]
    oldest_rbt = list(rbt_coll.aggregate(pick_oldest_rbt))
    return oldest_rbt


# list variables for which WXA should be kept

asm_keep = []
qc_keep = []
rbt_keep = []
asm_delete = []
qc_delete = []
rbt_delete = []
do_delete = []
test_list = list(asm_coll.aggregate(pipeline))


def data_objects_to_delete(output_list):
    for do in output_list:
        do_delete.append(do)


def find_asm_dups(omics_identifier, asm_keep_id):
    asm_duplicate_cursor = asm_coll.find(
        {"was_informed_by": omics_identifier, "id": {"$not": {"$regex": asm_keep_id}}}
    )
    for dup_asm in asm_duplicate_cursor:
        asm_delete.append(dup_asm["id"])
        data_objects_to_delete(dup_asm["has_output"])


def find_qc_dups(omics_identifier, qc_keep_id):
    qc_duplicate_cursor = qc_coll.find(
        {"was_informed_by": omics_identifier, "id": {"$not": {"$regex": qc_keep_id}}}
    )
    for qc_dup in qc_duplicate_cursor:
        qc_delete.append(qc_dup["id"])
        data_objects_to_delete(qc_dup["has_output"])


def find_rbt_dups(omics_identifier, rbt_keep_id):
    rbt_duplicate_cursor = rbt_coll.find(
        {"was_informed_by": omics_identifier, "id": {"$not": {"$regex": rbt_keep_id}}}
    )
    for rbt_dup in rbt_duplicate_cursor:
        rbt_delete.append(rbt_dup["id"])
        data_objects_to_delete(rbt_dup["has_output"])


for doc in test_list:
    omics_id = doc["_id"]
    records = pick_records(omics_id)
    if len(list(records)) == 1:
        for omics in records:
            if (
                len(omics["upstream_filtering"]) == 1
                and len(omics["rbt_record_to_keep"]) == 1
            ):
                asm_keep.append(omics["id"])
                find_asm_dups(doc["_id"], omics["id"])
                for filtering in omics["upstream_filtering"]:
                    qc_keep.append(filtering["id"])
                    find_qc_dups(doc["_id"], filtering["id"])
                for rbt in omics["rbt_record_to_keep"]:
                    rbt_keep.append(rbt["id"])
                    find_rbt_dups(doc["_id"], rbt["id"])
            elif (
                len(omics["upstream_filtering"]) == 1
                and len(omics["rbt_record_to_keep"]) == 0
            ):
                continue
            elif (
                len(omics["upstream_filtering"]) == 1
                and len(omics["rbt_record_to_keep"]) > 1
            ):
                # aggregation to get oldest filtering record
                asm_keep.append(omics["id"])
                for filtering in omics["upstream_filtering"]:
                    qc_keep.append(filtering["id"])
                    find_qc_dups(doc["_id"], filtering["id"])
                find_asm_dups(doc["_id"], omics["id"])
                input_list = omics["has_input"]
                target_do = input_list[0]
                oldest_rbt = pick_rbt(target_do)
                if len(list(oldest_rbt)) == 1:
                    for rbt in oldest_rbt:
                        rbt_keep.append(rbt["id"])
                    find_rbt_dups(doc["_id"], rbt["id"])
                else:
                    print("not sure what to do")
    else:
        print("not sure what to do for " + omics_id)

print(len(rbt_keep))
print(len(asm_keep))
print(len(qc_keep))

print(len(rbt_delete))
print(len(qc_delete))
print(len(asm_delete))
print(len(do_delete))


def make_deletes(delete_list, del_coll):
    request_body_file = del_coll + "_request_body.json"
    delete_statements = []
    for record in delete_list:
        delete_statements.append({"q": {"id": record}, "limit": 1})
    request_body_json = {"delete": del_coll, "deletes": delete_statements}
    with open(request_body_file, "w") as f:
        json.dump(request_body_json, f)


make_deletes(rbt_delete, rbt_coll_name)
make_deletes(qc_delete, qc_coll_name)
make_deletes(asm_delete, asm_coll_name)
make_deletes(do_delete, do_coll_name)



# need to match destination collection to api_request
# lots of redundancy
# prefer not to rewrite another paging method, so payloads could be large (but no mare than 2000)
# will there be a simple filter for all the collections to be queried?
# nmdc:sty-12-85j6kq06query from https://github.com/microbiomedata/nmdc-schema/issues/1205#issuecomment-1776129426 not working. try nmdc:sty-11-28tm5d36
# works: https://api-napa.microbiomedata.org/biosamples?filter=part_of%3Anmdc:sty-11-28tm5d36&per_page=999
# and https://api-napa.microbiomedata.org/biosamples?filter=part_of%3Anmdc:sty-11-28tm5d36&per_page=99&page=2
# may still require loosening of schema or "migrations" that assert missing data
import re

# 'collection_time': '08:45  - 12:05' !?
# study descriptions with stray whitespace

import click
import requests
import yaml


def dump_python_dict_to_yaml_file(python_dict, yaml_file_path):
    """Dumps a Python dict to a YAML file.

    Args:
      python_dict: The Python dict to dump.
      yaml_file_path: The path to the YAML file.
    """

    with open(yaml_file_path, "w") as yaml_file:
        yaml.dump(python_dict, yaml_file)


def read_python_dict_from_yaml_file(yaml_file_path):
    """Reads a Python dict from a YAML file.

    Args:
      yaml_file_path: The path to the YAML file.

    Returns:
      A Python dict.
    """

    with open(yaml_file_path, "r") as yaml_file:
        yaml_string = yaml_file.read()

    python_dict = yaml.load(yaml_string, Loader=yaml.SafeLoader)

    return python_dict


def extract_collection(string):
    """Extracts the text from a string after 'https://api-napa.microbiomedata.org/nmdcschema/' but before the first '?'.

    Args:
      string: The string to extract the text from.

    Returns:
      A string containing the extracted text.
    """

    regex = r'https:\/\/api-napa.microbiomedata.org\/nmdcschema\/(.*?)\?'
    match = re.search(regex, string)
    if match:
        return match.group(1)
    else:
        return None


@click.command()
@click.option("--api-url", multiple=True)
@click.option("--output-file", required=True, type=click.Path())
def main(output_file, api_url):
    database_dict = {}

    for current_url in api_url:
        collection_name = extract_collection(current_url)
        print(f"Populating the {collection_name} from {current_url}")
        if collection_name in database_dict:
            # append to existing list
            database_dict[collection_name].extend(requests.get(current_url).json()["resources"])
        else:
            database_dict[collection_name] = requests.get(current_url).json()["resources"]

    dump_python_dict_to_yaml_file(database_dict, output_file)


if __name__ == "__main__":
    main()



import json
import csv


def gen_delete(delete_list: list, key: str, limit: int, del_coll: str):
    r"""
    Generates HTTP request bodies that can be sent to the Runtime API's
    `/queries:run` endpoint in order to delete documents from the database.
    """
    if limit not in (0, 1):
        raise ValueError("limit must be 0 or 1.")
    delete_statement = []
    for record in delete_list:
        delete_statement.append({"q": {key: r"%s" % record}, "limit": limit})
    request_body_json = {"delete": del_coll, "deletes": delete_statement}
    request_body_file = del_coll + "_request_body.json"
    with open(request_body_file, "w") as f:
        json.dump(request_body_json, f)


# delete legacy orphan functional_annotation_agg records
delete_agg_list = []
with open("to_delete_nmdc.functional_annotation_agg.csv", newline="") as csvfile:
    agg_records = csv.reader(csvfile, delimiter=",")
    for row in agg_records:
        delete_agg_list.append(row[0])
del_coll = "functional_annotation_agg"
gen_delete(delete_agg_list, "metagenome_annotation_id", 0, del_coll)

# delete leagacy orphan data_object_set records
# uses regex as a workaround for newline issues double backslash issues
delete_do_list = []
with open("legacy_orphan.nmdc.data_object_set.csv", newline="") as csvfile:
    do_records = csv.reader(csvfile, delimiter=",")
    for row in do_records:
        strip_id = row[1].replace("\\n", "")
        print(strip_id)
        delete_do_list.append({"q": {"id": {"$regex": strip_id}}, "limit": 1})
    request_body_json = {"delete": "data_object_set", "deletes": delete_do_list}
    request_body_file = "data_object_set_request_body.json"
    with open(request_body_file, "w") as f:
        json.dump(request_body_json, f)



import pprint

from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper

schema_file = '../nmdc_schema/nmdc_materialized_patterns.yaml'

ignored_slots = [
    'aliases',
    'annotations',
    'comments',
    'examples',
    'maximum_cardinality',
    'minimum_cardinality',
    'multivalued',
    'name',
    'notes',
    'pattern',
    'range',
    'required',
    'structured_pattern',
    'title',
    'todos',
]  # todo which of these are reasonable as slot_usage?

schema_view = SchemaView(schema_file)

schema_classes = schema_view.all_classes()

for ck, cv in schema_classes.items():
    if "slot_usage" in cv and cv["slot_usage"]:
        su = cv["slot_usage"]
        for suk, suv in su.items():
            suv_dict = suv.__dict__
            for uk, uv in suv_dict.items():
                useful_dict = {}
                if uk not in ignored_slots and uv:
                    useful_dict[uk] = uv
                if useful_dict:
                    print(f"{suk} used in {ck}")
                    pprint.pprint(useful_dict)



import json
import os
import pprint
import re
from io import StringIO

import click
import curies
import pandas as pd
# import sparql_dataframe # copied and pasted, for addition of low_memory=False
from SPARQLWrapper import SPARQLWrapper, CSV, SELECT, POST, POSTDIRECTLY
from linkml_runtime import SchemaView

from SPARQLBurger.SPARQLQueryBuilder import *

pd.set_option('display.max_rows', None)
pd.set_option('display.max_columns', None)


# todo unpack bank node objects RECURSIVELY
# todo https://w3id.org/mixs/ or https://w3id.org/mixs-6-2-rc/ ?
# todo support for MULTIPLE constraints
# todo (manually) be on the lookout for "http" or GraphDB's default blank node prefix: "_:genid-" in the output (TSV)
#  suggests that more prefix definitions are needed
# todo may add a lot of prefixes that won't be used

# todo reuse code already in this script
# todo don't assume optional or multivalued


def replace_colons_and_whitespace(any_string):
    return re.sub(r"[: ]", "_", any_string)


def find_none_indices(list1):
    """
    Returns a list of the indices at which a Python list has a None value.

    Args:
      list1: The list to search.

    Returns:
      A list of the indices at which `list1` has a None value.
    """

    none_indices = []
    for i in range(len(list1)):
        if list1[i] is None:
            none_indices.append(i)

    return none_indices


def get_values_at_indices(list1, indices):
    """
    Returns the values in `list1` at the indices specified in `indices`.

    Args:
      list1: The list to get values from.
      indices: A list of indices.

    Returns:
      A list of the values in `list1` at the indices specified in `indices`.
    """

    values_at_indices = []
    for index in indices:
        values_at_indices.append(list1[index])

    return values_at_indices


def get_sparql_dataframe(endpoint, query, post=False):
    sparql = SPARQLWrapper(endpoint)
    sparql.setQuery(query)
    if sparql.queryType != SELECT:
        raise QueryException("Only SPARQL SELECT queries are supported.")

    if post:
        sparql.setOnlyConneg(True)
        sparql.addCustomHttpHeader("Content-type", "application/sparql-query")
        sparql.addCustomHttpHeader("Accept", "text/csv")
        sparql.setMethod(POST)
        sparql.setRequestMethod(POSTDIRECTLY)

    sparql.setReturnFormat(CSV)
    results = sparql.query().convert()
    _csv = StringIO(results.decode('utf-8'))
    return pd.read_csv(_csv, sep=",", low_memory=False)


class LinkMLClassSparqlQuery:
    def __init__(self,
                 concatenation_suffix,
                 do_not_query,
                 graph_name,  # not being used by sparqlburger ?
                 jsonld_context_jsons,
                 schema_file,
                 sparql_endpoint,
                 target_class_name,
                 target_p_o_constraint,
                 do_group_concat=False,  # not being used by sparqlburger ?
                 make_distinct=False,  # not being used by sparqlburger ?
                 ):
        self.schema_view = SchemaView(schema_file)

        self.class_slot_usage = {}
        self.concatenation_suffix = concatenation_suffix
        self.converter = None
        self.default_prefix = self.schema_view.schema.default_prefix
        self.do_group_concat = do_group_concat
        self.do_not_query = do_not_query
        self.graph_name = graph_name
        self.groupers = []
        self.make_distinct = make_distinct
        self.jsonld_context_jsons = jsonld_context_jsons
        self.schema_file = schema_file
        self.select_line = []
        self.slot_dict = {}
        self.target_class_name = target_class_name
        self.target_class_slots = self.schema_view.class_induced_slots(target_class_name)
        self.triple_patterns = []
        self.vars_to_select = []
        self.sparql_endpoint = sparql_endpoint
        self.target_p_o_constraint = target_p_o_constraint

    def get_class_slot_usage(self, class_name):
        # todo add prefixes
        #  loop over self.converter
        prefix_lines = []
        for pref_k, expansion_v in self.converter.prefix_map.items():
            prefix_lines.append(f"PREFIX {pref_k}: <{expansion_v}>")
        prefix_block = "\n".join(prefix_lines)
        class_slot_usage_query = f"""
        {prefix_block}
        select ?p (count(?{class_name}) as ?count) 
        where  {{ 
          ?{class_name} a nmdc:{class_name} ; 
          ?p ?o . 
        }}
        group by ?p
        """  # todo need better handling for class urls, classes names containing whitespace, etc. here
        df = get_sparql_dataframe(self.sparql_endpoint, class_slot_usage_query, post=True)
        p_list = df['p'].tolist()
        curie_list = [self.contract_url(p) for p in p_list]
        df['contracted'] = curie_list
        return df

    def contract_column(self, df, column_name):
        column_as_list = df[column_name].tolist()
        contracted_list = [self.contract_multi_urls(i) for i in column_as_list]
        df[column_name] = contracted_list
        # none_outputs = find_none_indices(contracted_list)
        # un_compacted_inputs = get_values_at_indices(column_as_list,
        #                                             none_outputs)  # todo oops they have been removed already # also removes blank node ids
        # # print(f"{column_name} {un_compacted_inputs = }")
        return df

    def contract_multi_urls(self, any_url):
        if type(any_url) == str:
            urls = any_url.split("|")
            contracted_urls = []
            for current_url in urls:
                contracted_url = self.contract_url(current_url)
                contracted_urls.append(contracted_url)
            cleaned_urls = [item for item in contracted_urls if item is not None]
            pasted_urls = "|".join(cleaned_urls)
            return pasted_urls
        else:
            return any_url

    def contract_url(self, any_url):
        try:
            curie = self.converter.compress(any_url)
            if curie:
                return curie
            else:
                return any_url
        except Exception as e:
            print(f"WARNING on {any_url}: {e}")  # todo logging not printing
            return any_url

    def populate_prefix_maps(self):

        # https://curies.readthedocs.io/en/latest/tutorial.html#loading-a-context

        # this could be simplified and can explicitly handle upper/lower case prefixes
        # https://curies.readthedocs.io/en/latest/tutorial.html#standardization
        # we're not using rdflib here, but for future reference
        # https://curies.readthedocs.io/en/latest/tutorial.html#integrating-with-rdflib
        # also
        # https://pypi.org/project/SPARQL-Burger/
        #   select_query.add_prefix

        for_chaining = []
        for jsonld_context_json in self.jsonld_context_jsons:
            print(f"Loading prefixes from {jsonld_context_json}")
            with open(jsonld_context_json) as json_file:
                temp_dict = json.load(json_file)
            if '@context' in temp_dict and '@vocab' in temp_dict['@context']:
                del temp_dict['@context']['@vocab']
            temp_context = curies.load_jsonld_context(temp_dict)
            for_chaining.append(temp_context)

        self.converter = curies.chain(for_chaining)

    def characterize_slot(self, slot_name, all_usage=None):
        local_schema_view = self.schema_view
        slot_obj = local_schema_view.induced_slot(slot_name, self.target_class_name)
        slot_range_type_name = type(local_schema_view.get_element(slot_obj.range)).class_name

        local_slot_dict = {
            "name": slot_obj.name,
            "slot_uri": slot_obj.slot_uri,
            "do_not_query": slot_obj.name in self.do_not_query,
            "multivalued": slot_obj.multivalued,
            "range": slot_obj.range,
            "range_type": slot_range_type_name,
            "required": slot_obj.required,
            "slot_slug": replace_colons_and_whitespace(slot_obj.name),
        }

        if slot_range_type_name == "class_definition":
            if local_schema_view.get_identifier_slot(slot_obj.range):
                local_slot_dict['class_identifier_slot'] = local_schema_view.get_identifier_slot(
                    slot_obj.range).name

        if not local_slot_dict['slot_uri']:
            local_slot_dict['slot_uri'] = f"{self.default_prefix}:{local_slot_dict['slot_slug']}"

        if all_usage is not None:
            slot_usage = all_usage[all_usage['contracted'].eq(local_slot_dict['slot_uri'])]
            if slot_usage.empty:
                local_slot_dict['do_not_query'] = True

        return local_slot_dict

    def characterize_slots(self):
        all_usage = self.class_slot_usage[self.target_class_name]
        for slot in self.target_class_slots:
            local_slot_dict = self.characterize_slot(slot.name, all_usage)
            self.slot_dict[slot.name] = local_slot_dict

    def create_triple(self, subject, slot_dict):
        triple = Triple(subject=subject, predicate=slot_dict["slot_uri"],
                        object=f"{subject}_{slot_dict['slot_slug']}")
        return triple

    def build_with_sparql_burger(self):
        scalar_groupers = []

        select_query = SPARQLSelectQuery(distinct=True, include_popular_prefixes=False)  # limit=1_000_000
        # Create an empty graph pattern
        main_pattern = SPARQLGraphPattern()
        # Add prefixes
        for k, v in self.converter.prefix_map.items():
            select_query.add_prefix(prefix=Prefix(prefix=k, namespace=v))

        # determine the IRI for the target class
        if self.schema_view.get_class(self.target_class_name).class_uri:
            type_query_object = self.schema_view.get_class(self.target_class_name).class_uri
        else:
            type_query_object = f"{self.default_prefix}:{self.target_class_name}"

        # determine variable for type query
        type_query_variable = f"?{replace_colons_and_whitespace(self.target_class_name)}"

        # add the typing BGP
        main_pattern.add_triples(
            triples=[
                Triple(subject=type_query_variable, predicate="rdf:type", object=type_query_object),
            ]
        )

        # Add the type variable for selecting
        select_query.add_variables(variables=[type_query_variable])
        scalar_groupers.append(type_query_variable)

        for sk, sv in self.slot_dict.items():
            if sv['do_not_query']:
                continue

            deeper_triples = []
            type_pattern = None
            if sv['range_type'] == "class_definition":
                # assume optional multivalued
                type_triple = Triple(subject=f"{type_query_variable}_{sv['slot_slug']}", predicate="rdf:type",
                                     object=f"{type_query_variable}_{sv['slot_slug']}_rdf_type")
                type_pattern = SPARQLGraphPattern(optional=True)
                type_pattern.add_triples(triples=[type_triple])
                if 'class_identifier_slot' not in sv:
                    deeper_slots = self.schema_view.class_induced_slots(sv['range'])
                    for ds in deeper_slots:
                        deeper_slot_dict = self.characterize_slot(ds.name)
                        deeper_triple = self.create_triple(f"{type_query_variable}_{sv['slot_slug']}", deeper_slot_dict)
                        deeper_triples.append(deeper_triple)

            triple = self.create_triple(type_query_variable, sv)

            if sv['required']:
                main_pattern.add_triples(triples=[triple])
                if type_pattern:
                    main_pattern.add_nested_graph_pattern(type_pattern)
                    select_query.add_variables(variables=[
                        f"(group_concat(distinct {type_query_variable}_{sv['slot_slug']}_rdf_type; SEPARATOR='|') as {type_query_variable}_{sv['slot_slug']}_rdf_type{self.concatenation_suffix})"])
                for deeper_triple in deeper_triples:
                    select_query.add_variables(variables=[
                        f"(group_concat(distinct {deeper_triple.object}; SEPARATOR='|') as {deeper_triple.object}{self.concatenation_suffix})"])
                    deeper_pattern = SPARQLGraphPattern(optional=True)
                    deeper_pattern.add_triples(triples=[deeper_triple])
                    main_pattern.add_nested_graph_pattern(deeper_pattern)
            else:
                optional_pattern = SPARQLGraphPattern(optional=True)
                optional_pattern.add_triples(triples=[triple])
                if type_pattern:
                    optional_pattern.add_nested_graph_pattern(type_pattern)
                    select_query.add_variables(variables=[
                        f"(group_concat(distinct {type_query_variable}_{sv['slot_slug']}_rdf_type; SEPARATOR='|') as {type_query_variable}_{sv['slot_slug']}_rdf_type{self.concatenation_suffix})"])
                for deeper_triple in deeper_triples:
                    select_query.add_variables(variables=[
                        f"(group_concat(distinct {deeper_triple.object}; SEPARATOR='|') as {deeper_triple.object}{self.concatenation_suffix})"])
                    deeper_pattern = SPARQLGraphPattern(optional=True)
                    deeper_pattern.add_triples(triples=[deeper_triple])
                    optional_pattern.add_nested_graph_pattern(deeper_pattern)
                main_pattern.add_nested_graph_pattern(optional_pattern)

            if deeper_triples:
                pass  # don't include the variable in the projection
            else:
                if sv['multivalued']:
                    select_query.add_variables(variables=[
                        f"(group_concat(distinct {type_query_variable}_{sv['slot_slug']}; SEPARATOR='|') as {type_query_variable}_{sv['slot_slug']}{self.concatenation_suffix})"])
                else:
                    select_query.add_variables(variables=[f"{type_query_variable}_{sv['slot_slug']}"])
                    scalar_groupers.append(f"{type_query_variable}_{sv['slot_slug']}")

        select_query.add_group_by(group=GroupBy(variables=scalar_groupers))

        if self.target_p_o_constraint:
            target_p_o_constraint_parts = self.target_p_o_constraint.split(" ")
            constraint_p = target_p_o_constraint_parts[0]
            constraint_o = target_p_o_constraint_parts[1]
            main_pattern.add_triples(triples=[
                Triple(subject=type_query_variable, predicate=constraint_p, object=constraint_o)])

        # Set the WHERE part
        select_query.set_where_pattern(graph_pattern=main_pattern)

        # write a multi-line string to a text file
        with open(f"{self.target_class_name}.rq", "w") as f:
            f.write(select_query.get_text())

        df = self.query_contract_drop(select_query.get_text())

        df.to_csv(f"{self.target_class_name}.tsv", index=False, sep="\t")

    def query_contract_drop(self, sparql_query, contract=True, drop_nas=True):
        df = get_sparql_dataframe(self.sparql_endpoint, sparql_query, post=True)
        if contract:
            for dfc in list(df.columns):
                df = self.contract_column(df, f"{dfc}")
        if drop_nas:
            df = df.dropna(axis=1, how='all')
        return df


class QueryException(Exception):
    pass


@click.command()  # todo make the file options paths, not strings
@click.option("--concatenation-suffix", default="s")
@click.option("--do-group-concat", is_flag=True)
@click.option("--do-not-query", multiple=True)
@click.option("--graph-name", default="mongodb://mongo-loadbalancer.nmdc.production.svc.spin.nersc.gov:27017")
@click.option("--make-distinct", is_flag=True)
@click.option(
    "-p",
    "--jsonld-context-jsons",
    multiple=True,
    required=False,
    type=click.Path(dir_okay=False),
)
@click.option("--query-file", help="Optional SPARQL query file. Bypasses class-based query building.")
@click.option("--schema-file", default="nmdc_schema/nmdc_materialized_patterns.yaml")
@click.option("--sparql-endpoint", default="https://graphdb-dev.microbiomedata.org/repositories/nmdc-metadata")
@click.option("--target-class-name", default="Study")
@click.option("--target-p-o-constraint",
              help="The predicate and literal object of an optional constraint on the target class")
def main(
        concatenation_suffix,
        do_group_concat,
        do_not_query,
        graph_name,
        make_distinct,
        query_file,
        schema_file,
        sparql_endpoint,
        target_class_name,
        target_p_o_constraint,
        jsonld_context_jsons,
):
    # SETUP
    query_support = LinkMLClassSparqlQuery(
        concatenation_suffix=concatenation_suffix,
        do_group_concat=do_group_concat,  # sparql burger mode assumes group concat
        do_not_query=do_not_query,
        graph_name=graph_name,  # sparql burger is not named graph aware
        make_distinct=make_distinct,  # sparql burger mode is not using this but could
        jsonld_context_jsons=jsonld_context_jsons,
        schema_file=schema_file,  # todo decouple schemaview construction for sake of bquery_file mode?
        sparql_endpoint=sparql_endpoint,
        target_class_name=target_class_name,
        target_p_o_constraint=target_p_o_constraint,
        # todo allow for multiple constraints, even with a subject other than the target class?
    )

    query_support.populate_prefix_maps()

    if query_file:
        print(f"query_file = {query_file}")
        with open(query_file, "r") as f:
            sparql_query = f.read()
        df = query_support.query_contract_drop(sparql_query)
        basename = os.path.basename(query_file)
        output_file_name = basename.replace(".rq", ".tsv")
        df.to_csv(output_file_name, index=False, sep="\t")
        exit()

    query_support.class_slot_usage[target_class_name] = query_support.get_class_slot_usage(target_class_name)

    query_support.characterize_slots()

    query_support.build_with_sparql_burger()


if __name__ == "__main__":
    main()



from typing import List
from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "7.8.0"
    _to_version = "8.0.0"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("biosample_set", [self.standardize_letter_casing_of_gold_biosample_identifiers])
        self.adapter.process_each_document("extraction_set", [self.rename_sample_mass_field])
        self.adapter.process_each_document("omics_processing_set", [self.standardize_letter_casing_of_gold_sequencing_project_identifiers])
        self.adapter.process_each_document("study_set", [self.standardize_letter_casing_of_gold_study_identifier])

    def rename_sample_mass_field(self, extraction: dict) -> dict:
        r"""
        Renames the `sample_mass` field to `input_mass`.

        >>> m = Migrator()
        >>> m.rename_sample_mass_field({'id': 123})  # no `sample_mass` field
        {'id': 123}
        >>> m.rename_sample_mass_field({'id': 123, 'sample_mass': 456})  # test: renames field and preserves value
        {'id': 123, 'input_mass': 456}
        """

        self.logger.info(f"Starting migration of {extraction['id']}")
        if "sample_mass" in extraction:
            extraction["input_mass"] = extraction.pop("sample_mass")
        return extraction

    def standardize_letter_casing_of_gold_identifiers(self, identifiers: List[str]) -> List[str]:
        r"""
        Replaces the prefix `GOLD:` with `gold:` in the list of identifiers.

        >>> m = Migrator()
        >>> m.standardize_letter_casing_of_gold_identifiers(['GOLD:prefix_was_uppercase'])
        ['gold:prefix_was_uppercase']
        >>> m.standardize_letter_casing_of_gold_identifiers(['gold:prefix_was_lowercase'])
        ['gold:prefix_was_lowercase']
        >>> m.standardize_letter_casing_of_gold_identifiers(['Gold:prefix_remains_mixed_case'])
        ['Gold:prefix_remains_mixed_case']

        Note: If the identifier contains more than one colon, everything after and including
              the second colon will be discarded.

        >>> m.standardize_letter_casing_of_gold_identifiers(['GOLD:preserves_this:but_not_this'])
        ['gold:preserves_this']
        """

        standardized_identifiers = []
        for identifier in identifiers:
            self.logger.info(f"Original identifier: {identifier}")
            curie_parts = identifier.split(":")
            curie_prefix = curie_parts[0]  # everything before the first colon
            # everything after the first colon, assuming there are no more colons
            curie_local_id = curie_parts[1]

            # Note: `s.split(":", maxsplit=1)` could be used to divide the string at the
            #       _first_ colon only, so that `parts[1]` contains everything after that.
            #       See: https://docs.python.org/3/library/stdtypes.html#str.split

            if curie_prefix == "GOLD":
                standardized_identifiers.append(f"gold:{curie_local_id}")
            else:
                standardized_identifiers.append(identifier)

        return standardized_identifiers

    def standardize_letter_casing_of_gold_biosample_identifiers(self, biosample: dict) -> dict:
        r"""
        Converts uppercase "GOLD:" prefixes into lowercase "gold:" prefixes, in `gold_biosample_identifiers` values.

        >>> m = Migrator()
        >>> m.standardize_letter_casing_of_gold_biosample_identifiers({'id': 123})
        {'id': 123, 'gold_biosample_identifiers': []}
        >>> m.standardize_letter_casing_of_gold_biosample_identifiers(
        ...     {'id': 123, 'gold_biosample_identifiers': ['GOLD:foo', 'gold:bar']}
        ... )
        {'id': 123, 'gold_biosample_identifiers': ['gold:foo', 'gold:bar']}
        """

        field_name = "gold_biosample_identifiers"
        if field_name in biosample and biosample[field_name]:
            biosample[field_name] = self.standardize_letter_casing_of_gold_identifiers(biosample[field_name])
        else:
            biosample[field_name] = []
        return biosample

    def standardize_letter_casing_of_gold_sequencing_project_identifiers(self, omics_processing: dict) -> dict:
        r"""
        Converts uppercase "GOLD:" prefixes in `gold_sequencing_project_identifiers` values into lowercase "gold:".

        >>> m = Migrator()
        >>> m.standardize_letter_casing_of_gold_sequencing_project_identifiers({'id': 123})
        {'id': 123, 'gold_sequencing_project_identifiers': []}
        >>> m.standardize_letter_casing_of_gold_sequencing_project_identifiers(
        ...     {'id': 123, 'gold_sequencing_project_identifiers': ['GOLD:foo', 'gold:bar']}
        ... )
        {'id': 123, 'gold_sequencing_project_identifiers': ['gold:foo', 'gold:bar']}
        """

        field_name = "gold_sequencing_project_identifiers"
        if field_name in omics_processing and omics_processing[field_name]:
            omics_processing[field_name] = self.standardize_letter_casing_of_gold_identifiers(omics_processing[field_name])
        else:
            omics_processing[field_name] = []
        return omics_processing

    def standardize_letter_casing_of_gold_study_identifier(self, study: dict) -> dict:
        r"""
        Converts uppercase "GOLD:" prefixes in `gold_study_identifiers` values into lowercase "gold:".

        >>> m = Migrator()
        >>> m.standardize_letter_casing_of_gold_study_identifier({'id': 123})
        {'id': 123, 'gold_study_identifiers': []}
        >>> m.standardize_letter_casing_of_gold_study_identifier(
        ...     {'id': 123, 'gold_study_identifiers': ['GOLD:foo', 'gold:bar']}
        ... )
        {'id': 123, 'gold_study_identifiers': ['gold:foo', 'gold:bar']}
        """

        field_name = "gold_study_identifiers"
        if field_name in study and study[field_name]:
            study[field_name] = self.standardize_letter_casing_of_gold_identifiers(study[field_name])
        else:
            study[field_name] = []
        return study



from copy import deepcopy
from typing import Optional, Callable, List, Union
from nmdc_schema.migrators.adapters.adapter_base import AdapterBase


class DictionaryAdapter(AdapterBase):
    r"""
    Class containing methods related to manipulating a "database" represented by
    a Python dictionary.
    """

    def __init__(self, database: dict, **kwargs) -> None:
        r"""
        Invokes the initialization method of the parent class, passing to it any specified keyword arguments.
        Also initializes the reference to the database this adapter instance will be used to manipulate.
        """
        super().__init__(**kwargs)
        self._db = database

    def create_collection(self, collection_name: str) -> None:
        r"""
        Creates an empty collection having the specified name, if no collection by that name exists.
        Also invokes `self.on_collection_created`, if defined, passing to it the name of the collection.

        References:
        - https://docs.python.org/3/library/functions.html#callable

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.create_collection("thing_set")
        >>> "thing_set" in database
        True
        >>> len(database["thing_set"])  # existing collection will retain existing contents
        3
        >>> da.create_collection("item_set")
        >>> "item_set" in database
        True
        >>> len(database["item_set"])  # new collection will be empty
        0
        """
        if collection_name not in self._db:
            self._db[collection_name] = []

            # If the relevant callback function exists, invoke it.
            if callable(self.on_collection_created):
                self.on_collection_created(collection_name)

    def rename_collection(self, current_name: str, new_name: str) -> None:
        r"""
        Renames the specified collection, if it exists, so that it has the specified new name.
        Also invokes `self.on_collection_renamed`, if defined, passing to it the old and new names of the collection.

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.rename_collection("thing_set", "item_set")
        >>> "thing_set" in database
        False
        >>> "item_set" in database
        True
        """
        if current_name in self._db:
            self._db[new_name] = self._db.pop(current_name)

            # If the relevant callback function exists, invoke it.
            if callable(self.on_collection_renamed):
                self.on_collection_renamed(current_name, new_name)

    def delete_collection(self, collection_name: str) -> None:
        r"""
        Deletes the collection having the specified name, if such a collection exists.
        Also invokes `self.on_collection_deleted`, if defined, passing to it the name of the collection.

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.delete_collection("thing_set")
        >>> "thing_set" in database
        False
        """
        if collection_name in self._db:
            del self._db[collection_name]

            # If the relevant callback function exists, invoke it.
            if callable(self.on_collection_deleted):
                self.on_collection_deleted(collection_name)

    def insert_document(self, collection_name: str, document: dict) -> None:
        r"""
        Inserts the specified document into the collection having the specified name.

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.insert_document("thing_set", {"id": "444", "foo": "dee"})
        >>> database["thing_set"][-1]  # gets last item in list
        {'id': '444', 'foo': 'dee'}
        """
        self._db[collection_name].append(document)

    def get_document_having_value_in_field(
        self, collection_name: str, field_name: str, value: str
    ) -> Optional[dict]:
        r"""
        Retrieves the first document from the specified collection, having the specified value in the specified field.

        Note: This only supports top-level fields (e.g. `_id`, `depth`), not nested fields (e.g. `depth.has_unit`).

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.get_document_having_value_in_field("thing_set", "id", "222")
        {'id': '222', 'foo': 'baz'}
        >>> da.get_document_having_value_in_field("thing_set", "foo", "baz")
        {'id': '222', 'foo': 'baz'}
        >>> da.get_document_having_value_in_field("thing_set", "id", "no_such_value") is None
        True
        >>> da.get_document_having_value_in_field("thing_set", "no_such_field", "111") is None
        True
        """
        try:
            # Create a generator over which we can iterate via `next`.
            #
            # Note: A generator is a special type of iterator. We create it here via a so-called "generator expression".
            #       You can think of a "generator expression" as a list comprehension that doesn't involve loading the
            #       entire sequence into memory.
            #
            # Reference: https://docs.python.org/3/howto/functional.html#generator-expressions-and-list-comprehensions
            #
            document_generator = (
                d for d in self._db[collection_name] if d.get(field_name) == value
            )

            # Return the first document yielded by the generator (or `None` if the generator doesn't yield one).
            document = next(document_generator, None)
        except KeyError:
            document = None

        return document

    def get_document_having_one_of_values_in_field(
        self, collection_name: str, field_name: str, values: List[str]
    ) -> Optional[dict]:
        r"""
        Retrieves the first document from the specified collection, having any one of the specified values in the
        specified field.

        Note: This only supports top-level fields (e.g. `_id`, `depth`), not nested fields (e.g. `depth.has_unit`).

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.get_document_having_one_of_values_in_field("thing_set", "id", ["221", "222", "223"])
        {'id': '222', 'foo': 'baz'}
        >>> da.get_document_having_one_of_values_in_field("thing_set", "foo", ["baa", "baz", "bab"])
        {'id': '222', 'foo': 'baz'}
        >>> da.get_document_having_one_of_values_in_field("thing_set", "id", ["no_such_value"]) is None
        True
        >>> da.get_document_having_one_of_values_in_field("thing_set", "foo", []) is None  # no values to match with
        True
        """
        try:
            # Create a generator over which we can iterate via `next`.
            document_generator = (
                d for d in self._db[collection_name] if d.get(field_name) in values
            )

            # Return the first document yielded by the generator (or `None` if the generator doesn't yield one).
            document = next(document_generator, None)
        except KeyError:
            document = None

        return document

    def delete_documents_having_value_in_field(
        self, collection_name: str, field_name: str, value: str
    ) -> int:
        r"""
        Deletes all documents from the specified collection, having the specified value in the specified field;
        and returns the number of documents that were deleted.

        Note: This only supports top-level fields (e.g. `_id`, `depth`), not nested fields (e.g. `depth.has_unit`).

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "222", "foo": "blue"},  # same id, so that we can...
        ...     {"id": "222", "foo": "blue"},  # ...test deleting multiple.
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> len(database["thing_set"])
        5
        >>> da.delete_documents_having_value_in_field("thing_set", "id", "no_such_value")
        0
        >>> da.delete_documents_having_value_in_field("thing_set", "no_such_field", "111")
        0
        >>> len(database["thing_set"])
        5
        >>> da.delete_documents_having_value_in_field("thing_set", "id", "222")  # deletes 3 documents
        3
        >>> len(database["thing_set"])
        2
        >>> da.delete_documents_having_value_in_field("thing_set", "foo", "qux")
        1
        >>> len(database["thing_set"])
        1
        """
        # Filter out the documents having the specified value in the specified field.
        # Reference: https://docs.python.org/3/library/functions.html#filter
        documents_initial = self._db[collection_name]
        document_generator = filter(
            lambda d: field_name not in d or d.get(field_name) != value,
            documents_initial,
        )
        documents_remaining = list(document_generator)

        # Update the collection so that it consists of the filtered result.
        self._db[collection_name] = documents_remaining

        # Return the number of documents that were deleted.
        num_documents_deleted = len(documents_initial) - len(documents_remaining)
        return num_documents_deleted

    def process_each_document(
        self, collection_name: str, pipeline: List[Callable[[dict], dict]]
    ) -> None:
        r"""
        Passes each document in the specified collection through the specified processing pipeline—in which
        the output of any given function is the input to the function after it—and stores the final output
        back in the collection, replacing the original document.

        Reference: https://docs.python.org/3/library/copy.html#copy.deepcopy

        >>> def capitalize_foo_value(thing: dict) -> dict:
        ...     thing["foo"] = thing["foo"].upper()
        ...     return thing
        >>>
        >>> database = {
        ...   "thing_set": [
        ...     {"id": "111", "foo": "bar"},
        ...     {"id": "222", "foo": "baz"},
        ...     {"id": "333", "foo": "qux"}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.process_each_document("thing_set", [capitalize_foo_value])
        >>> database["thing_set"][0]
        {'id': '111', 'foo': 'BAR'}
        >>> database["thing_set"][1]
        {'id': '222', 'foo': 'BAZ'}
        >>> da.process_each_document("missing_set", [capitalize_foo_value])  # non-existent collection does not trigger an exception
        """

        # Iterate over every document in the collection, if the collection exists.
        if collection_name in self._db:
            for index, original_document in enumerate(self._db[collection_name]):
                # Make a copy of the original document.
                #
                # Note: This isn't technically necessary (we could modify the original document in place
                #       while it resides in the Python array), but this keeps the algorithm analogous to
                #       what I expect its "real database" (e.g. MongoDB) counterparts to be.
                #
                processed_document = deepcopy(original_document)

                # "Pass" the document through the functions (i.e. "stages") that make up the pipeline,
                # such that the output from one stage becomes the input to the next stage.
                for function in pipeline:
                    processed_document = function(processed_document)

                # Overwrite the original document with the processed one.
                self._db[collection_name][index] = processed_document

    def set_field_of_each_document(
        self, collection_name: str, field_name: str, value: Union[None, str, int, float, bool],
    ) -> None:
        r"""
        Assigns the specified value to the specified field of each document in the collection.
        This method is a specialized alternative to the `process_each_document` method.

        >>> database = {
        ...   "thing_set": [
        ...     {"id": "1", "name": "original"},
        ...     {"id": "2"},
        ...     {"id": "3", "name": None}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> da.set_field_of_each_document("thing_set", "name", "new")
        >>> database["thing_set"][0]
        {'id': '1', 'name': 'new'}
        >>> database["thing_set"][1]
        {'id': '2', 'name': 'new'}
        >>> database["thing_set"][2]
        {'id': '3', 'name': 'new'}
        """

        # Iterate over every document in the collection, if the collection exists.
        if collection_name in self._db:
            for document in self._db[collection_name]:
                document[field_name] = value

    def do_for_each_document(
        self, collection_name: str, action: Callable[[dict], None]
    ) -> None:
        r"""
        Passes each document in the specified collection to the specified function. This method was designed
        to facilitate iterating over all documents in a collection without actually modifying them.

        >>> total = 0
        >>> def add_to_total(payment: dict) -> None:
        ...     global total
        ...     total += payment["amount"]
        >>>
        >>> database = {
        ...   "payment_set": [
        ...     {"id": "111", "amount": 100},
        ...     {"id": "222", "amount": 200},
        ...     {"id": "333", "amount": 300}
        ...   ]
        ... }
        >>> da = DictionaryAdapter(database)
        >>> total
        0
        >>> da.do_for_each_document("payment_set", add_to_total)
        >>> total
        600
        """

        # Iterate over every document in the collection, if the collection exists.
        if collection_name in self._db:
            for document in self._db[collection_name]:
                action(document)
